// Contains class definitions
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
#include "RawCD.h"
#include <iostream.h>


// ***************
//  Here, we define the actual functions



//||||||||||||||||||||||||||||||||||||||||||||||||
//
//  Class RawCD
//
//||||||||||||||||||||||||||||||||||||||||||||||||



//.......................................
//
// Inlined private methods
//
//.......................................


// Converts a single msf-structure to an integer value in units of
// frames.
//
inline
int RawCD::msf2int(struct cdrom_msf0 msf)
{
  return (msf.frame + CD_FRAMES*(msf.second + 60*msf.minute));
}


// Converts a single integer value in units of frames to an
// msf-structure.
//
inline
struct cdrom_msf0 RawCD::int2msf(int tt)
{
  struct cdrom_msf0 out;
  out.minute = tt/60/CD_FRAMES;
  out.second = (tt/CD_FRAMES)%60;
  out.frame = tt%CD_FRAMES;
  return out;
}


// Converts a pair of integers in units of frames to an msf-block.
//
inline
struct cdrom_msf RawCD::int2msf_block(int t0, int tf)
{
  struct cdrom_msf out;
  out.cdmsf_min0 = t0/60/CD_FRAMES;
  out.cdmsf_sec0 = (t0/CD_FRAMES)%60;
  out.cdmsf_frame0 = t0%CD_FRAMES;
  out.cdmsf_min1 = tf/60/CD_FRAMES;
  out.cdmsf_sec1 = (tf/CD_FRAMES)%60;
  out.cdmsf_frame1 = tf%CD_FRAMES;
  return out;
}


// Checks for ioctl errors;
//
inline 
bool RawCD::ioctlOk(int stat) 
{
  if(stat != 0)
	{		 
	perror("ioctl");
	playerError = true;
	return false;
	}
  else
	{
	playerError = false;
	return true;
	}
}


//.....
//
// Methods
//
// Constructor for the RawCD object is inlined.
// No declaration needed.
//......


// Update the status flags and counter variables.  Does nothing if the
// tray is ejected or empty.
//
void RawCD::UpdateStatus(void)
{
  int stat;
  struct cdrom_subchnl cd_stat;

  if( isEjected | isEmpty )
	return;

  cd_stat.cdsc_format = CDROM_MSF;
  stat = ioctl(devfp,CDROMSUBCHNL,&cd_stat);
  if( ioctlOk(stat) )
	{
	switch(cd_stat.cdsc_audiostatus)
	   {
	    case CDROM_AUDIO_PLAY:
		  {
		  isPlaying = true;
		  isStopped = isPaused = false;
		  break;
		  }

	    case CDROM_AUDIO_PAUSED:
		  {
		  isPlaying = isPaused = true;
		  isStopped = false;
		  break;
		  }

	    case CDROM_AUDIO_COMPLETED:
		  {
		  isPlaying = isPaused = false;
		  isStopped = playComplete = true;
		  break;
		  }
		
	    case CDROM_AUDIO_NO_STATUS:
		  {
		  break;
		  }
		
	    default:
		  {
		  playerError = isStopped = true;
		  isPlaying = isPaused = false;
		  break;
		  }
	   }

	citem = cd_stat.cdsc_trk;
	abs_ctime = msf2int(cd_stat.cdsc_absaddr.msf);
	rel_ctime = abs_ctime - toc[citem];
	}
  else //ioctl trouble
	{
	if(errno == NO_MEDIA_PRESENT)
	   {
	   isEmpty = true;
	   isPlaying = isStopped = isPaused = false;
	   return;
	   }
	else
	   perror(device_file_name);
	}
}


// Uses the device filename [set by its overloaded version or by the
// constructor]  to open the CDROM device, then loads the ToC, and
// lastly calls \Ref{UpdateStatus}().
//
void RawCD::InitDev(void)
{
  int stat,i,j;
  struct cdrom_tocentry cdtoc;
  struct cdrom_tochdr cdhead;
  struct cdrom_volctrl cdv;


  // Default value of the error flag.
  playerError = false;

  if( (devfp == -1) || isEjected )
	{
	// need to open the device.  assume it auto-closes the tray.
	devfp = open(device_file_name,O_RDONLY);
	isEmpty = isEjected = false;
	if(devfp == -1)
	   {
	   if(errno == NO_MEDIA_PRESENT)
		 {
		 isEmpty = true;
		 isPlaying = isStopped = isPaused = false;
		 return;
		 }
	   else
		 perror(device_file_name);
	   }
	}
  
  // Get the header info
  stat = ioctl(devfp,CDROMREADTOCHDR,&cdhead);
  if( !(ioctlOk(stat)) )
	{
	isEmpty = true;
	isPlaying = isStopped = isPaused = false;
	playerError = true;
	return;
	}

  // find the number of tracks.
  first_item = cdhead.cdth_trk0;
  last_item = cdhead.cdth_trk1;
  nitems = last_item  - first_item + 1; 
  if(nitems <= 0)
	{
	isEmpty = true;
	isPlaying = isStopped = isPaused = false;
	return;
	}
#ifdef DEBUG_ON
  cerr << "%dbg%\t";
  cerr << "N=" << nitems << "\tFirst=" << first_item << "\tLast="
	  << last_item << hex << endl;
  cerr << "%dbg%";
#endif

  //Add extra space for the leadout track and the non-existent '0' track.
  toc = new int[nitems+2];

  // Now loop and get the info on each track.
  isEmpty = true;
  cdtoc.cdte_format = CDROM_MSF;
  for(i=first_item, j=cdhead.cdth_trk0; i<=last_item; i++, j++)
	{
	cdtoc.cdte_track = j;
	stat = ioctl(devfp,CDROMREADTOCENTRY,&cdtoc);
	if( ioctlOk(stat) && !(cdtoc.cdte_ctrl & CDROM_DATA_TRACK) )
	   {
	   isEmpty = false;
	   toc[i] = msf2int(cdtoc.cdte_addr.msf);
	   }
	else 	// the toc is zero by default.
	   playerError = true;
#ifdef DEBUG_ON
	cerr << "\ttoc[" << i << "]=0x" << toc[i] << endl;
#endif
	}

  //Next we get the leadout track entry, which tells us how long the
  //entire cd is.
  cdtoc.cdte_track = CDROM_LEADOUT;
  stat = ioctl(devfp,CDROMREADTOCENTRY,&cdtoc);
  if( ioctlOk(stat) )
	{
	toc[last_item+1] = msf2int(cdtoc.cdte_addr.msf);
#ifdef DEBUG_ON
	cerr << "\ttoc[" << i << "] = len = 0x" << toc[i] 
		<< dec << endl;
#endif
	}
  else
	playerError = true;

  // Lastly, get the current volume level and do a status update.
  stat = ioctl(devfp,CDROMVOLREAD,&cdv);
  if( ioctlOk(stat) )
	{
	vol = (float)cdv.channel0/255.0;  // vol = [0,1]
	}
  playComplete = isPlaying = isStopped = isPaused = false;
  UpdateStatus();
#ifdef DEBUG_ON
  cerr << "%dbg%\t";
  cerr << "vol=" << vol << endl;
  cerr << "%dbg%\t";
  cerr << "flags: " << isStopped << " " << isPlaying << " " 
	  << isPaused << " " << isEmpty << " " << playComplete << " " 
	  << playerError << " " << isEjected << endl;
#endif
}


// Ejects the CD.
//
void RawCD::Eject(void)
{
  int stat;

  if(devfp == -1)
	isEjected = true; // Force a re-init
  
  if( isEjected )
	return;

  if( isPlaying || isPaused )
	Stop();

  stat = ioctl(devfp,CDROMEJECT);
  if( ioctlOk(stat) )
	{
	isStopped = false;
	isEjected = true;
	isEmpty = true;
	playerError = false;
	}
  else
	playerError = true;
}


// Closes the tray by one of two methods, depending on how the
// routine is compiled.  It can use the CDROMCLOSETRAY {\tt ioctl}, or
// it can simply close and reopen the CDROM device.  It always calls
// \Ref{InitDev}(void) after everything else. 
//
void RawCD::CloseTray(void)
{
  int stat;

  // Check some flags to figure out what to do.
  if(!noTray)
	{
	if(hasCloseIoctl)
	   {
	   stat = ioctl(devfp,CDROMCLOSETRAY);
	   if( ioctlOk(stat) )
		 isEjected = false;
	   else
		 {
		 cerr << "Your CDROM Drive seems to have a "
			 << "broken driver or does" << endl;
		 cerr << "not support the ioctl to close the tray.  "
			 << "If your drive" << endl;
		 cerr << "*does* automatically close "
			 << "the tray when you mount CD's" << endl;
		 cerr << "normally, try turning off the close-tray-ioctl." 
			 << endl;
		 }
	   }
	else // !hasCloseIoctl
	   {
	   close(devfp);
	   isEjected = true;    // Causes InitDev() to do a reopen.
	   }
	}
  // Always do this.
  InitDev();
}


// Increments the volume by a step.
//
void RawCD::IncVol(void)
{
  int stat;
  struct cdrom_volctrl cdv;

  if( isEmpty )
	{
	return;
	}

  vol += volstep;
  if(vol > 1.0)
	{
	vol = 1.0;
	}

  cdv.channel0 = (u_char)(255*vol);
  cdv.channel1 = cdv.channel2 = cdv.channel3 = cdv.channel0;
  stat = ioctl(devfp,CDROMVOLCTRL,&cdv);
  ioctlOk(stat);
}


// Decrements the volume by a step.
//
void RawCD::DecVol(void)
{
  int stat;
  struct cdrom_volctrl cdv;

  if( isEmpty )
	{
	return;
	}

  vol -= volstep;
  if(vol < 0.0)
	{
	vol = 0.0;
	}

  cdv.channel0 = (u_char)(255*vol);
  cdv.channel1 = cdv.channel2 = cdv.channel3 = cdv.channel0;
  stat = ioctl(devfp,CDROMVOLCTRL,&cdv);
  ioctlOk(stat);
}


// Stops play, updating the device status and setting internal flags.
// Only does something if the drive is closed and has a disc.
//
void RawCD::Stop(void)
{
  int stat;

  if( isEjected | isEmpty )
	{
	playerError = false;
	return;
	}
  else
	{
	UpdateStatus();
	stat = ioctl(devfp,CDROMSTOP);
	if( ioctlOk(stat) )
	   {
	   isPlaying = false;
	   isPaused = false;
	   isStopped = true;
	   playComplete = false;
	   playerError = false;
	   lrqpitem = 0;
	   }
	else
	   playerError = true;
	}
}


// Toggles the pause state of the cdrom.  Only does something if the
// drive is playing.
//
void RawCD::PauseResume(void)
{
  int stat;

  if( !isPlaying )
	{
#ifdef DEBUG_ON
	cerr << "%dbg%\tPauseReq-NOP" << endl;
#endif
	isPaused = false;
	playerError = false;
	return;
	}
  else if( isPaused )
	{
#ifdef DEBUG_ON
	cerr << "%dbg%\tPauseReq - resume" << endl;
#endif
	stat = ioctl(devfp,CDROMRESUME);
	if( ioctlOk(stat) )
	   {
	   isPaused = false;
	   playerError = false;
	   }
	else
	   playerError = true;
	}
  else // !isPaused
	{
#ifdef DEBUG_ON
	cerr << "%dbg%\tPauseReq - pause" << endl;
#endif
	stat = ioctl(devfp,CDROMPAUSE);
	if( ioctlOk(stat) )
	   {
	   isPaused = true;
	   playerError = false;
	   }
	else
	   playerError = true;
	}
}


// Play track {\em itemno}.  {\em itemno}=0 corresponds to the first
// playable track.
//
// The flag {\em reposition} is used to control what {\bf
// PlayItem} does if the drive is already playing.  If {\em
// reposition=true}, the drive stops and restarts at the new
// location.  Otherwise, {\bf PlayItem} does nothing.
//
// This is where the lrpq-index comes in.  It specifies the index of
// the last specifically-requested play item, and is set by both {\bf
// PlayItem} and \Ref{PlayItemBlock}.  If:
// \begin{enumerate}
// \item {\em reposition=true}
// \item the requested track is the one immediately following the last
// one requested [{\it i.e. itemno=lrqp+1}].
// \item the current play position is +/-1 second of the beginning of
// {\em itemno}
// \end{enumerate}
// ...then {\bf PlayItem} just updates some counters and returns,
// allowing the drive to keep playing.  If the last two conditions
// aren't met, then the drive stops an restarts.
//
// The idea behind this elaborate scheme is to keep play smooth.  As
// long as your CDROM drive keeps playing through, in sequence, once
// it's started playing, this scheme will work.
// 
void RawCD::PlayItem(int itemno, bool reposition)
{
  int stat;
  struct cdrom_ti playrange_t;


  // intemno=0 <==> the first item.
  itemno += first_item;

#ifdef DEBUG_ON
  cerr << "%dbg%\tPlayReq: item #" << itemno << ", reposn=" 
	  << reposition << " [lrqp#=" << lrqpitem << "]" << endl;
#endif

  // Some checks.
  if( isPlaying && reposition )
	{
	UpdateStatus();
	// We want to check and see if the requested item is the next
	// item on the disc following the last item played.
	if (itemno == lrqpitem+1)
	   {
#ifdef DEBUG_ON
	   cerr << "%dbg%\tPlayReq: abs_t_diff=" 
		   << (toc[itemno]-abs_ctime);
#endif
	   // If we're within a second of either side of the new track
	   // starting point, do nothing except update some variables.
	   if( (toc[itemno]-CD_FRAMES <= abs_ctime) 
		  && (abs_ctime <= toc[itemno]+CD_FRAMES) )
		 {
#ifdef DEBUG_ON
		 cerr << "\t\t:NOP" << endl;
#endif
		 playComplete = isPaused = isStopped = false;
		 isPlaying = true;
		 lrqpitem = itemno;
		 return;
		 }
#ifdef DEBUG_ON
	   cerr << endl;
#endif
	   }

	// This happens if we're not at the end of track or the new track
	// is not the next in sequence.  Prevents infinite loops if the
	// player just stops before we reach the end of the track.
	Stop();
	}
  else if( isPlaying || isEmpty )
	{
	return;
	}

  if( (itemno < first_item) || (itemno > last_item) )
	{
	return;
	}

  // Set some status info.
  playComplete = isPaused = isStopped = false;
  isPlaying = true;

  //Now start play and get the time.
  playrange_t.cdti_trk0 = itemno;
  playrange_t.cdti_trk1 = last_item;
  playrange_t.cdti_ind0 = playrange_t.cdti_ind1 = 0;
  stat = ioctl(devfp,CDROMPLAYTRKIND,&playrange_t);
  if( !ioctlOk(stat) )
	{
	isPlaying = false;
	playerError = true;
	}
  else
	{
	lrqpitem = itemno;
	playerError = false;
	}
}


// Play track {\em itemno} between two blocks of time, starting at
// {\em t0} and ending at {\em t_f}.  {\em t0} and {\em t_f} are
// relative to the start of the specified track.  {\em itemno}=0
// corresponds to the first playable track.
//
// If {\em t_f}<={\em t0}, {\bf PlayItemBlock} returns an error.  If
// {\em t_f} is 0, negative, or beyond the end of the CD, {\bf
// PlayItemBlock} simply plays to the end of the disc.
//
// The flag {\em reposition} is used to control what {\bf
// PlayItemBlock} does if the drive is already playing.  If {\em
// reposition=true}, the drive stops and restarts at the new
// location.  Otherwise, {\bf PlayItemBlock} does nothing.
//
void RawCD::PlayItemBlock(int itemno, int t0, int t_f, bool reposition)
{
  int stat;
  struct cdrom_msf playrange;

  // intemno=0 <==> the first item.
  itemno += first_item;

#ifdef DEBUG_ON
  cerr << "%dbg%\tPlayReq: item #" << itemno << ", t_0=" << t0
	  << ", t_f=" << t_f << ", reposn=" << reposition << endl;
#endif

  // Some checks.
  if( isPlaying && reposition )
	{
	Stop();
	}
  else if( isPlaying || isEmpty )
	{
	return;
	}

  if( (itemno < first_item) || (itemno > last_item) )
	{
	return;
	}

  // Set some status info.
  playComplete = isPaused = isStopped = false;
  isPlaying = true;
  // Don't forget to offset t0 by the item's position on the disc!
  if( (t_f <= 0) || (t_f >= toc[last_item+1]) )
	{
	t_f = toc[last_item+1] - 1;
	}
  else
	{
	t_f += toc[itemno];
	}
  t0 += toc[itemno];

  if(t_f <= t0)
	{
	isPlaying = false;
	playerError = true;
	return;
	}

#ifdef DEBUG_ON
  cerr << "%dbg%\tPlayBlock: t_0=" << t0
	  << ", t_f=" << t_f << endl;
#endif

  //Now start play and get the time.
  playrange = int2msf_block(t0, t_f);
  stat = ioctl(devfp,CDROMPLAYMSF,&playrange);
  if( !ioctlOk(stat) )
	{
	isPlaying = false;
	playerError = true;
	}
  else
	{
	lrqpitem = itemno;
	playerError = false;
	}
}



//||||||||||||||||||||||||||||||||
//
// End
//
